/*****************************************************************************/
/* TAnchors.cpp                        Copyright Ladislav Zezula 2001 - 2005 */
/*---------------------------------------------------------------------------*/
/* A class for easy object anchoring. Those who compare the Delphi program-  */
/* ming with Visual C++, will certainly miss the "Anchors" property in visual*/
/* design. This class supports something like that, although it's not so     */
/* comfort like the visual design in Delphi/C++ Builder                      */
/*---------------------------------------------------------------------------*/
/*   Date    Ver   Who  Comment                                              */
/* --------  ----  ---  -----------------------------------------------------*/
/* 19.12.01  1.00  Lad  Created                                              */
/* 21.12.01  1.01  Lad  Added TRACE outputs for debug support                */
/* 28.12.01  1.02  Lad  Added the __USE_FAST_REDRAW__ flag                   */
/* 09.01.01  1.02  Lad  Fixed ugly effect caused by the group boxes.         */
/* 15.07.03  1.03  Lad  Added method using DeferWindowPos                    */
/* 31.08.04  2.00  Lad  Added support for proportional resizing              */
/*****************************************************************************/

#define __UTILS_SELF
#include "Utils.h"
#include "TAnchors.h"

// Choose one flag only !!!
//#define __USE_SET_WINDOW_POS__        // Redraws the windows using SetWindowPos + WS_CLIPCHILDREN
#define __USE_DEFER_WINDOW_POS__        // Redraws windows using DeferWindowPos 
//#define __USE_FAST_REDRAW__           // Redraws windows using double buffering

//-----------------------------------------------------------------------------
// Constructor and destructor

TAnchors::TAnchors()
{
    // Initialize anchors to NULL
    m_pFirstAnchor = NULL;
    m_hWndParent   = NULL;
    ZeroMemory(&m_rectParent, sizeof(RECT));
}

TAnchors::~TAnchors()
{
    TAnchor * pAnchor = m_pFirstAnchor;
    TAnchor * temp;

    // Free all used anchors
    while(pAnchor != NULL)
    {
        temp = pAnchor->pNext;
        delete pAnchor;
        pAnchor = temp;
    }
}

//-----------------------------------------------------------------------------
// Public functions

// This function adds the anchors for an object into the list of anchors.
// Later after the parent window has been resized, the TAnchors::AnchorAllObjects
// must be called to rearrange the window
BOOL TAnchors::AddAnchor(HWND hWnd, DWORD dwAnchors)
{
    TAnchor * pNewAnchor;
    HWND      hParent = GetParent(hWnd);// Parent of given window
    RECT      rectParent;               // Rectangle of the parent
    RECT      rectWnd;                  // Window (child) rect
    int       nParentWidth;
    int       nParentHeight;

    // Test if the window passed is valid
    if(IsWindow(hWnd) == FALSE || IsWindow(hParent) == FALSE)
    {
        OutputDebugString(_T("Invalid window handle passed to TAnchors::AddAnchor\n"));
        return FALSE;
    }

    // Add the parent window
    if(m_hWndParent == NULL)
        m_hWndParent = hParent;

#ifdef __USE_SET_WINDOW_POS__
    AddWsClipChildren(hParent)
#endif  // __USE_SET_WINDOW_POS__

    // Create the new anchor
    pNewAnchor = new TAnchor;
    ZeroMemory(pNewAnchor, sizeof(TAnchor));
    pNewAnchor->hWnd      = hWnd;

    // Retrieve the screen-relative rectangle of the parent's client area
    GetParentWindowRect(hParent, rectParent);
    nParentWidth  = rectParent.right - rectParent.left;
    nParentHeight = rectParent.bottom - rectParent.top;

    // Get the screen-relative rectangle of the child window
    GetWindowRect(hWnd, &rectWnd);

    // Store the relative position and distance from left side of the parent
    pNewAnchor->nLeftRel = akNotAnchored;
    if(dwAnchors & akLeft)
    {
        pNewAnchor->nLeftRel   = 0;
        pNewAnchor->nLeftSpace = rectWnd.left - rectParent.left;
    }
    if(dwAnchors & akLeftCenter)
    {
        pNewAnchor->nLeftRel   = 50;
        pNewAnchor->nLeftSpace = rectWnd.left - (rectParent.left + (nParentWidth / 2));
    }

    // Store the relative position and distance from top side of the parent
    pNewAnchor->nTopRel = akNotAnchored;
    if(dwAnchors & akTop)
    {
        pNewAnchor->nTopRel   = 0;
        pNewAnchor->nTopSpace = rectWnd.top - rectParent.top;
    }
    if(dwAnchors & akTopCenter)
    {
        pNewAnchor->nTopRel   = 50;
        pNewAnchor->nTopSpace = rectWnd.top - (rectParent.top + (nParentHeight / 2));
    }

    // Store the relative position and distance from right side of the parent
    pNewAnchor->nRightRel = akNotAnchored;
    if(dwAnchors & akRight)
    {
        pNewAnchor->nRightRel   = 100;
        pNewAnchor->nRightSpace = rectParent.right - rectWnd.right;
    }
    if(dwAnchors & akRightCenter)
    {
        pNewAnchor->nRightRel   = 50;
        pNewAnchor->nRightSpace = (rectParent.left + (nParentWidth / 2)) - rectWnd.right;
    }

    // Store the relative position and distance from bottom side of the parent
    pNewAnchor->nBottomRel = akNotAnchored;
    if(dwAnchors & akBottom)
    {
        pNewAnchor->nBottomRel   = 100;
        pNewAnchor->nBottomSpace = rectParent.bottom - rectWnd.bottom;
    }
    if(dwAnchors & akBottomCenter)
    {
        pNewAnchor->nBottomRel   = 50;
        pNewAnchor->nBottomSpace = (rectParent.top + (nParentHeight / 2)) - rectWnd.bottom;
    }

    // Insert anchor in the list of anchors
    InsertAnchor(pNewAnchor);
    return TRUE;
}

BOOL TAnchors::AddAnchor(HWND hDlg, UINT nIDCtrl, DWORD dwAnchors)
{
    return AddAnchor(GetDlgItem(hDlg, nIDCtrl), dwAnchors);
}


BOOL TAnchors::AddAnchorEx(HWND hWnd, int nLeftRel, int nTopRel, int nRightRel, int nBottomRel)
{
    TAnchor * pNewAnchor;
    HWND      hParent = GetParent(hWnd);// Parent of given window
    RECT      rectParent;               // Rectangle of the parent
    RECT      rectWnd;                  // Window (child) rect
    int       nParentWidth;
    int       nParentHeight;

    // Test if the window passed is valid
    if(IsWindow(hWnd) == FALSE || IsWindow(hParent) == FALSE)
    {
        OutputDebugString(_T("Invalid window handle passed to TAnchors::AddAnchorEx\n"));
        return FALSE;
    }

    // Add the parent window
    if(m_hWndParent == NULL)
        m_hWndParent = hParent;

#ifdef __USE_SET_WINDOW_POS__
    AddWsClipChildren(hParent)
#endif  // __USE_SET_WINDOW_POS__

    // Create the new anchor
    pNewAnchor = new TAnchor;
    ZeroMemory(pNewAnchor, sizeof(TAnchor));
    pNewAnchor->hWnd      = hWnd;

    // Retrieve the screen-relative rectangle of the parent's client area
    GetParentWindowRect(hParent, rectParent);
    nParentWidth  = rectParent.right - rectParent.left;
    nParentHeight = rectParent.bottom - rectParent.top;

    // Get the screen-relative rectangle of the child window
    GetWindowRect(hWnd, &rectWnd);

    // Store the relative position and distance from left side of the parent
    if((pNewAnchor->nLeftRel = nLeftRel) != akNotAnchored)
    {
        int nLeftAbs = rectParent.left + (nParentWidth * pNewAnchor->nLeftRel) / 100;

        pNewAnchor->nLeftSpace = rectWnd.left - nLeftAbs;
    }

    // Store the relative position and distance from top side of the parent
    if((pNewAnchor->nTopRel = nTopRel) != akNotAnchored)
    {
        int nTopAbs = rectParent.top + (nParentHeight * pNewAnchor->nTopRel) / 100;

        pNewAnchor->nTopSpace = rectWnd.top - nTopAbs;
    }

    // Store the relative position and distance from right side of the parent
    if((pNewAnchor->nRightRel = nRightRel) != akNotAnchored)
    {
        int nRightAbs = rectParent.left + (nParentWidth * pNewAnchor->nRightRel) / 100;

        pNewAnchor->nRightSpace = nRightAbs - rectWnd.right;
    }

    // Store the relative position and distance from bottom side of the parent
    if((pNewAnchor->nBottomRel = nBottomRel) != akNotAnchored)
    {
        int nBottomAbs = rectParent.top + (nParentHeight * pNewAnchor->nBottomRel) / 100;

        pNewAnchor->nBottomSpace = nBottomAbs - rectWnd.bottom;
    }

    // Insert anchor in the list of anchors
    InsertAnchor(pNewAnchor);
    return TRUE;
}

BOOL TAnchors::AddAnchorEx(HWND hDlg, UINT nIDCtrl, int nLeftRel, int nTopRel, int nRightRel, int nBottomRel)
{
    return AddAnchorEx(GetDlgItem(hDlg, nIDCtrl), nLeftRel, nTopRel, nRightRel, nBottomRel);
}

// Sets the size of the parent window manually
// Use this function when the size of the parent is not final
// in the moment when the anchors are being created.
// (E.g. in MFC application when creating a view).
void TAnchors::SetParentRectManual(RECT & rectParent)
{
    m_rectParent.left   = 
    m_rectParent.top    = 0;
    m_rectParent.right  = rectParent.right - rectParent.left;
    m_rectParent.bottom = rectParent.bottom - rectParent.top;
}

// This function makes the complete anchoring. It must be called in response to
// WM_SIZE message, or whenever the child windows need to be repositioned.
BOOL TAnchors::AnchorAllObjects()
{
    TAnchor * pAnchor;
    RECT NewRect;                      // New position and size of the resized window

    // If no anchored windows, or the parent is not a valid window, do nothing
    if(m_pFirstAnchor == NULL || IsWindow(m_hWndParent) == FALSE)
        return FALSE;

#ifdef __USE_SET_WINDOW_POS__
    for(pAnchor = m_pFirstAnchor; pAnchor != NULL; pAnchor = pAnchor->pNext)
    {
        if(GetNewWindowRect(m_hWndParent, pAnchor, NewRect)) 
        {
            SetWindowPos(pAnchor->hWnd, NULL, NewRect.left, NewRect.top, NewRect.right, NewRect.bottom, SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOOWNERZORDER);
            InvalidateRect(pAnchor->hWnd, NULL, TRUE);
        }
    }
#endif

// This method uses xxxDeferWindowPos and is the best one. Deferred window
// positioning is a part of Windows API and thus supported by Microsoft.
#ifdef __USE_DEFER_WINDOW_POS__
    HDWP  hDwp = NULL;
    int   nWindows = 0;

    // Count the anchored windows and begin the window positioning
    for(pAnchor = m_pFirstAnchor; pAnchor != NULL; pAnchor = pAnchor->pNext)
        nWindows++;
    hDwp = BeginDeferWindowPos(nWindows);
    for(pAnchor = m_pFirstAnchor; pAnchor != NULL; pAnchor = pAnchor->pNext)
    {
        if(GetNewWindowRect(m_hWndParent, pAnchor, NewRect)) 
            hDwp = DeferWindowPos(hDwp, pAnchor->hWnd, NULL, NewRect.left, NewRect.top, NewRect.right, NewRect.bottom, SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOOWNERZORDER);
    }
    EndDeferWindowPos(hDwp);
#endif

// This method is similar to using SetWindowPos, except that the redrawing
// is completely directed by us.
#ifdef __USE_FAST_REDRAW__
    SendMessage(m_hWndParent, WM_SETREDRAW, FALSE, 0);
    for(pAnchor = m_pFirstAnchor; pAnchor != NULL; pAnchor = pAnchor->pNext)
    {
        if(GetNewWindowRect(m_hWndParent, pAnchor, NewRect)) 
            SetWindowPos(pAnchor->hWnd, NULL, NewRect.left, NewRect.top, NewRect.right, NewRect.bottom, SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOOWNERZORDER);
    }

    // Fast redraw logic does not work properly in this moment
    // Problem : how to redraw the non-client area ?
    SendMessage(m_hWndParent, WM_SETREDRAW, TRUE, 0);
    PaintWindow(m_hWndParent, NULL);
#endif // __USE_FAST_REDRAW__

    return TRUE;
}

//-----------------------------------------------------------------------------
// Function operating with TAnchor object

void TAnchors::InsertAnchor(TAnchor * pNewAnchor)
{
    TAnchor * pLastAnchor = m_pFirstAnchor;

    if(m_pFirstAnchor == NULL)
        m_pFirstAnchor = pNewAnchor;
    else
    {
        while(pLastAnchor->pNext != NULL)
            pLastAnchor = pLastAnchor->pNext;

        pLastAnchor->pNext = pNewAnchor;
    }
}

void TAnchors::GetParentWindowRect(HWND hParent, RECT & rect)
{
    // If the parent rect has been set manually,
    // use this setting
    if(m_rectParent.right != 0 && m_rectParent.bottom != 0)
        rect = m_rectParent;
    else
        GetClientRect(hParent, &rect);

    // Convert the rectangle to screen coordinates
    ClientToScreen(hParent, (LPPOINT)&rect.left);
    ClientToScreen(hParent, (LPPOINT)&rect.right);
}

//-----------------------------------------------------------------------------
// This function calculates the new position of the window after the size
// of its parent has been changed. The returned coordinates ar stored
// in the NewRect as following:
//
//  - NewRect.left   - New X position of the child window (relative to parent)
//  - NewRect.top    - New Y position of the child window (relative to parent)
//  - NewRect.right  - New width of the child window
//  - NewRect.bottom - New height of the child window

BOOL TAnchors::GetNewWindowRect(HWND hParent, TAnchor * pAnchor, RECT & NewRect)
{
    RECT  rectParent;                   // Client rectangle of the parent window
                                        // (after it's been resized)
    RECT  rectWnd;                      // Screen rectangle of the child window
    int   nParentWidth;
    int   nParentHeight;
    int   nWndWidth;
    int   nWndHeight;
    int   x, y, cx, cy;                 // New window pos and size

    // Retrieve the client dimensions
    GetClientRect(hParent, &rectParent);
    nParentWidth  = rectParent.right;
    nParentHeight = rectParent.bottom;
    ClientToScreen(hParent, (LPPOINT)&rectParent.left);
    ClientToScreen(hParent, (LPPOINT)&rectParent.right);

    GetWindowRect(pAnchor->hWnd, &rectWnd);
    nWndWidth  = rectWnd.right - rectWnd.left;
    nWndHeight = rectWnd.bottom - rectWnd.top;

    // If the window has not been destroyed and is a child of the given one
    if(IsWindow(pAnchor->hWnd) && GetParent(pAnchor->hWnd) == hParent)
    {
        // Calculate the left-top-right-bottom positions of the child window
        x = rectWnd.left - rectParent.left;
        if(pAnchor->nLeftRel != akNotAnchored)
        {
            int nLeftAbs = (nParentWidth * pAnchor->nLeftRel) / 100;

            x = nLeftAbs + pAnchor->nLeftSpace;
        }

        y = rectWnd.top - rectParent.top;
        if(pAnchor->nTopRel != akNotAnchored)
        {
            int nTopAbs = (nParentHeight * pAnchor->nTopRel) / 100;

            y = nTopAbs + pAnchor->nTopSpace;
        }

        cx = nWndWidth;
        if(pAnchor->nRightRel != akNotAnchored)
        {
            int nRightAbs = (nParentWidth * pAnchor->nRightRel) / 100;

            if(pAnchor->nLeftRel == akNotAnchored)
                x = nRightAbs - pAnchor->nRightSpace - nWndWidth;

            cx = (nRightAbs - pAnchor->nRightSpace) - x;
        }

        cy = nWndHeight;
        if(pAnchor->nBottomRel != akNotAnchored)
        {
            int nBottomAbs = (nParentHeight * pAnchor->nBottomRel) / 100;

            if(pAnchor->nTopRel == akNotAnchored)
                y = nBottomAbs - pAnchor->nBottomSpace - nWndHeight;

            cy = (nBottomAbs - pAnchor->nBottomSpace) - y;
        }
   
        // Store the result into the rectangle
        NewRect.left   = x;
        NewRect.top    = y;
        NewRect.right  = cx;
        NewRect.bottom = cy;
        return TRUE;
    }

    return FALSE;
}

// Add the WS_CLIPCHILDREN style to the parent. This fastens the window
// redrawing. But I have found that the groupboxes make very ugly effect due to their
// bad redrawing logic. So we must be careful and add the style only if no
// groupbox was found inside the window.
void TAnchors::AddWsClipChildren(HWND hParent)
{
    BOOL bFoundGroupBox = FALSE;

    // Search all children. If we find a groupbox, 
    // we cannot add the style
    for(HWND hChild = GetWindow(hParent, GW_CHILD); hChild != NULL; hChild = GetWindow(hChild, GW_HWNDNEXT))
    {
        char  szClassName[64];
        DWORD dwStyle = GetWindowLong(hChild, GWL_STYLE);

        // If we knew the group box, stop the searching (one group box is enough)
        GetClassNameA(hChild, szClassName, sizeof(szClassName));
        if(!stricmp(szClassName, "Button") && (dwStyle & BS_GROUPBOX))
        {
            bFoundGroupBox = TRUE;
            break;
        }
    }

    // Add the WS_CLIPCHILDREN style
    if(bFoundGroupBox == FALSE)
        SetWindowLong(hParent, GWL_STYLE, GetWindowLong(hParent, GWL_STYLE) | WS_CLIPCHILDREN);
}

//-----------------------------------------------------------------------------
// Fast-window repaint logic. These function were written because the default
// window repaint login caused ugly flicking effects when resizing window with
// anchored objects.

// This function moves the window origin by dx and dy pixels
void TAnchors::MoveWindowOrg(HDC hDC, int dx, int dy)
{
    POINT p;

    GetWindowOrgEx(hDC, &p);
    SetWindowOrgEx(hDC, p.x - dx, p.y - dy, NULL);        
}

// This function retrieves the widths of both borders (X, Y). It was implemented because
// the "GetWindowInfo()" function returns the wrong sizes when WS_EX_CLIENTEDGE and
// WS_EX_STATICEDGE are set.
void TAnchors::GetBorderSize(HWND hWnd, POINT * border)
{
    DWORD dwExStyle = GetWindowLong(hWnd, GWL_EXSTYLE);

    border->x = border->y = 0;
    if(dwExStyle & WS_EX_CLIENTEDGE)
    {
        border->x += 2;
        border->y += 2;
    }
    if(dwExStyle & WS_EX_STATICEDGE)
    {
        border->x ++;
        border->y ++;
    }
    if(GetWindowLong(hWnd, GWL_STYLE) & WS_BORDER)
    {
        border->x ++;
        border->y ++;
    }
}

// This function repaints the window frame. It was not tested for dialogs
// and overlapped windows, but it's normally used only for dialog childs.
// This function was written because the window borders were not
// repainted on some windows (e.g. progress box).
void TAnchors::PaintFrame(HWND hWnd, HDC hDC, RECT * edgeRect)
{
    DWORD dwExStyle = GetWindowLong(hWnd, GWL_EXSTYLE);

    // Draw the client edge, if any
    if(dwExStyle & WS_EX_CLIENTEDGE)
        DrawEdge(hDC, edgeRect, EDGE_SUNKEN, BF_RECT | BF_ADJUST);

    // Draw the static edge, if any
    if(dwExStyle & WS_EX_STATICEDGE)
        DrawEdge(hDC, edgeRect, BDR_SUNKENOUTER, BF_RECT | BF_ADJUST);

    // Draw normal border, if any
    if(GetWindowLong(hWnd, GWL_STYLE) & WS_BORDER)
        DrawEdge(hDC, edgeRect, BDR_OUTER, BF_RECT | BF_ADJUST | BF_MONO);
/*
    GetClassName(hWnd, className, _tsizeof(className));
    if(!_tcsicmp(className, _T("Edit")))
    {
        HBRUSH hBrush;
        RECT   clientRect;

        GetClientRect(hWnd, &clientRect);
        clientRect.left++; clientRect.top++;
        hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNSHADOW));
        FrameRect(hDC, &clientRect, hBrush);
        DeleteObject(hBrush);

        clientRect.left = clientRect.top = 0;
        clientRect.right++; clientRect.bottom++;
        hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNHIGHLIGHT));
        FrameRect(hDC, &clientRect, hBrush);
        DeleteObject(hBrush);
    }
*/
}

// This function repaints all window children (e.g. dialog box controls) into
// a specified device context.
BOOL TAnchors::PaintChildrenTo(HWND hWnd, HDC hDC)
{
    POINT border;
    POINT pt;
    RECT  rect;
    int   nSaveIndex;

    // Get left-top window position
    pt.x = pt.y = 0;
    ClientToScreen(hWnd, &pt);

    for(HWND hChild = GetWindow(hWnd, GW_CHILD); hChild != NULL; hChild = GetWindow(hChild, GW_HWNDNEXT))
    {
        if(IsWindowVisible(hChild) == FALSE)
            continue;

        // Retrieve the child position and dimension
        GetWindowRect(hChild, &rect);

        // Save the DC state and prepare it for child drawing
        nSaveIndex = SaveDC(hDC);
        MoveWindowOrg(hDC, rect.left - pt.x, rect.top - pt.y);
        SetRect(&rect, 0, 0, rect.right - rect.left, rect.bottom - rect.top);
        IntersectClipRect(hDC, 0, 0, rect.right, rect.bottom);

        // Draw the window content. We have to calculate with the border
        GetBorderSize(hChild, &border);
        MoveWindowOrg(hDC, border.x, border.y);
        PaintWindow(hChild, hDC);

        // Now draw the frame
        MoveWindowOrg(hDC, -border.x, -border.y);
        PaintFrame(hChild, hDC, &rect);

        RestoreDC(hDC, nSaveIndex);
    }
    return TRUE;
}

// This function paints a window into a specific context. If the context is NULL,
// the function draws the window into window device context using "double buffering"
// (Whe whole window and its controls are first painted into a memory DC and then copied
// into the window DC).
BOOL TAnchors::PaintWindowTo(HWND hWnd, HDC hDC)
{
    // Paint the window
    SendMessage(hWnd, WM_ERASEBKGND, (WPARAM)hDC, 0);
    SendMessage(hWnd, WM_PAINT, (WPARAM)hDC, 0);
    return PaintChildrenTo(hWnd, hDC);
}

BOOL TAnchors::PaintWindow(HWND hWnd, HDC hDC)
{
    HBITMAP hBmp;                       // Compatible bitmap
    RECT    clientRect;                 // Parent's client rectangle
    HDC     hMemDC;                     // Memory DC

    // If already has a paint DC, draw it.
    if(hDC != NULL)
        return PaintWindowTo(hWnd, hDC);

    // Create a compatible bitmap with the size of the client rect
    GetClientRect(hWnd, &clientRect);
    hDC  = GetDC(NULL);                 // Device context of the entire screen
    hBmp = CreateCompatibleBitmap(hDC, clientRect.right, clientRect.bottom);
    ReleaseDC(NULL, hDC);
    if(hBmp == NULL)
        return FALSE;
    
    // Create the compatible DC (Initially with the size of 1x1 pixel) and select the bitmap.
    hMemDC = CreateCompatibleDC(NULL);
    hBmp = (HBITMAP)SelectObject(hMemDC, hBmp);

    // Start painting
    hDC = GetDC(hWnd);
    IntersectClipRect(hMemDC, 0, 0, clientRect.right, clientRect.bottom);
    PaintWindow(hWnd, hMemDC);

    // Copy the memory DC into the window DC
    BitBlt(hDC, 0, 0, clientRect.right, clientRect.bottom, hMemDC, 0, 0, SRCCOPY);
    ReleaseDC(hWnd, hDC);

    // Cleanup
    hBmp = (HBITMAP)SelectObject(hMemDC, hBmp);
    DeleteDC(hMemDC);
    DeleteObject(hBmp);
    return TRUE;
}
