/*****************************************************************************/
/* TShellCtxMenu.cpp                      Copyright (c) Ladislav Zezula 2004 */
/*---------------------------------------------------------------------------*/
/* Description:                                                              */
/*---------------------------------------------------------------------------*/
/*   Date    Ver   Who  Comment                                              */
/* --------  ----  ---  -------                                              */
/* 01.07.04  1.00  Lad  The first version of TShellContextMenu.cpp           */
/*****************************************************************************/

#define __UTILS_SELF
#include "Utils.h"
#include "TShellCtxMenu.h"

//-----------------------------------------------------------------------------
// Definitions

#define MIN_ID 1
#define MAX_ID 10000

//-----------------------------------------------------------------------------
// Global variables

static IContextMenu3 * g_pContextMenu3 = NULL;
static IContextMenu2 * g_pContextMenu2 = NULL;

//-----------------------------------------------------------------------------
// Constructor and destructor

TShellContextMenu::TShellContextMenu()
{
    m_pShellFolder = NULL;
    m_hHeap        = GetProcessHeap();
    m_ppFileNames  = NULL;
    m_dwFiles      = 0;
}

TShellContextMenu::~TShellContextMenu()
{
    if(m_ppFileNames != NULL)
    {
        for(DWORD i = 0; i < m_dwFiles; i++)
        {
            if(m_ppFileNames[i] != NULL)
                HeapFree(m_hHeap, 0, m_ppFileNames[i]);
        }
        HeapFree(m_hHeap, 0, m_ppFileNames);
        m_ppFileNames = NULL;
    }
}

//-----------------------------------------------------------------------------
// Public functions

// Adds a file from the file name.
int TShellContextMenu::AddFile(const TCHAR * szFileName)
{
    OLECHAR ** ppNewNames;
    OLECHAR * szFileNameW = NULL;
    size_t nLength = _tcslen(szFileName);

    szFileNameW = (OLECHAR *)HeapAlloc(m_hHeap, HEAP_ZERO_MEMORY, (nLength + 1) * sizeof(WCHAR));

#ifndef _UNICODE
    MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szFileName, nLength, szFileNameW, nLength);
#else
    _tcscpy(szFileNameW, szFileName);
#endif

    // Reallocate the PIDL list and insert the new item
    ppNewNames = (OLECHAR **)HeapAlloc(m_hHeap, HEAP_ZERO_MEMORY, sizeof(OLECHAR *) * (m_dwFiles + 1));
    if(m_ppFileNames != NULL)
        CopyMemory(ppNewNames, m_ppFileNames, (m_dwFiles * sizeof(OLECHAR *)));
    ppNewNames[m_dwFiles++] = szFileNameW;
    if(m_ppFileNames != NULL)
        HeapFree(m_hHeap, 0, m_ppFileNames);
    m_ppFileNames = ppNewNames;
    return ERROR_SUCCESS;
}

// Runs the context menu. Before calling this function, the CoInitialize
// must have been called for the current thread
int TShellContextMenu::RunMenu(HWND hWndParent, POINT pt)
{
    IContextMenu * pContextMenu = NULL;
    IShellFolder * pDesktopFolder = NULL;
    LPITEMIDLIST pidlItem = NULL;        // relative pidl
    LPITEMIDLIST pidl = NULL;
    WNDPROC OldWndProc = NULL;
    IMalloc * pMalloc = NULL;
    DWORD dwVersion = 0;
    HMENU hMenu = CreatePopupMenu();
    int nError = ERROR_SUCCESS;

    // If no files, do nothing
    if(m_dwFiles == 0)
        return ERROR_SUCCESS;

    // Retrieve the IShellFolder interface necessary for work
    if(nError == ERROR_SUCCESS)
        nError = SHGetDesktopFolder(&pDesktopFolder);
    if(nError == ERROR_SUCCESS)
        nError = SHGetMalloc(&pMalloc);

    // Allocate the space for the array of LPITEMIDLISTs
    if(nError == ERROR_SUCCESS)
    {
        m_ppItemIDList = (LPITEMIDLIST *)HeapAlloc(m_hHeap, HEAP_ZERO_MEMORY, m_dwFiles * sizeof(LPITEMIDLIST));
        if(m_ppItemIDList == NULL)
            nError = GetLastError();
    }

    // ParseDisplayName creates a PIDL from a file system path relative to the IShellFolder interface
    // but since we use the Desktop as our interface and the Desktop is the namespace root
    // that means that it's a fully qualified PIDL, which is what we need
    if(nError == ERROR_SUCCESS)
        nError = pDesktopFolder->ParseDisplayName(hWndParent, 0, m_ppFileNames[0], NULL, &pidl, NULL);
    
    if(nError == ERROR_SUCCESS)
    {
        nError = SHBindToParentEx(pidl, IID_IShellFolder, (void **)&m_pShellFolder, NULL);
        pMalloc->Free(pidl);
        pidl = NULL;
    }

    // now we have the IShellFolder interface to the parent folder specified in the first element in strArray
    // since we assume that all objects are in the same folder (as it's stated in the MSDN)
    // we now have the IShellFolder interface to every objects parent folder

    if(nError == ERROR_SUCCESS)
    {
        IShellFolder * pShellFolder = NULL;

        for(DWORD i = 0; i < m_dwFiles && nError == ERROR_SUCCESS; i++)
        {
            // Get the PIDL of the file name
            if(nError == ERROR_SUCCESS)
                nError = pDesktopFolder->ParseDisplayName(hWndParent, 0, m_ppFileNames[i], NULL, &pidl, NULL);
            
            // Bind to the parent PIDL
            if(nError == ERROR_SUCCESS)
                nError = SHBindToParentEx(pidl, IID_IShellFolder, (void **)&pShellFolder, (LPCITEMIDLIST *)&pidlItem);

            // Insert the PIDL into our array of PIDLs.
            if(nError == ERROR_SUCCESS)
                m_ppItemIDList[i] = CopyPIDL(pidlItem);

            if(pidlItem != NULL)
                HeapFree(m_hHeap, 0, pidlItem);
            if(pidl != NULL)
                pMalloc->Free(pidl);
            if(pShellFolder != NULL)
                pShellFolder->Release();
            pShellFolder = NULL;
            pidlItem = NULL;
            pidl = NULL;
        }
    }

    // Now retrieve the IContextMenu interface (as higher version as possible)
    if(nError == ERROR_SUCCESS)
        nError = GetIContextMenu(hWndParent, &pContextMenu, &dwVersion);

    // Lets fill our popupmenu  
    if(nError == ERROR_SUCCESS)
    {
        pContextMenu->QueryContextMenu(hMenu,
                                       GetMenuItemCount(hMenu),
                                       MIN_ID,
                                       MAX_ID,
                                       CMF_NORMAL | CMF_EXPLORE);
    }

    // If the version of IContextMenu is higher than 1,
    // subclass the window to have possibility to handle menu-related messages
    if(nError == ERROR_SUCCESS && dwVersion > 1)
    {
        // Set the global pointers to higher versions of the context menu
        if(dwVersion > 1)
            g_pContextMenu2 = (IContextMenu2 *)pContextMenu;
        if(dwVersion > 2)
            g_pContextMenu3 = (IContextMenu3 *)pContextMenu;

        OldWndProc = (WNDPROC)SetWindowLongPtr(hWndParent, GWLP_WNDPROC, (LONG_PTR)HookWndProc);
    }

    // Track the popup menu and invoke command, if any
    if(nError == ERROR_SUCCESS)
    {
        CMINVOKECOMMANDINFO cmi;
        UINT_PTR nIDCommand = TrackPopupMenu(hMenu, TPM_RETURNCMD | TPM_LEFTALIGN, pt.x, pt.y, 0, hWndParent, NULL);

        if(MIN_ID <= nIDCommand && nIDCommand <= MAX_ID)
        {
            ZeroMemory(&cmi, sizeof(CMINVOKECOMMANDINFO));
            cmi.cbSize = sizeof (CMINVOKECOMMANDINFO);
            cmi.lpVerb = (const char *)(nIDCommand - MIN_ID);
            cmi.nShow = SW_SHOWNORMAL;
            
            pContextMenu->InvokeCommand(&cmi);
        }
    }

    // Free the interfaces and unsubclass
    if(OldWndProc != NULL)
        SetWindowLongPtr(hWndParent, GWLP_WNDPROC, (LONG_PTR)OldWndProc);
    if(pContextMenu != NULL)
        pContextMenu->Release();
    if(m_pShellFolder != NULL)
        m_pShellFolder->Release();
    if(pDesktopFolder != NULL)
        pDesktopFolder->Release();
    if(pMalloc != NULL)
        pMalloc->Release();
    if(m_ppItemIDList != NULL)
        FreePIDLList(m_ppItemIDList, m_dwFiles);
    g_pContextMenu3 = NULL;
    g_pContextMenu2 = NULL;
    m_pShellFolder  = NULL;
    return ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// Protected variables

LRESULT CALLBACK TShellContextMenu::HookWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    { 
        case WM_MENUCHAR:    // Only supported by IContextMenu3
            if (g_pContextMenu3 != NULL)
            {
                LRESULT lResult = 0;
                g_pContextMenu3->HandleMenuMsg2 (uMsg, wParam, lParam, &lResult);
                return (lResult);
            }
            break;

        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            if(wParam != 0) 
                break; // if wParam != 0 then the uMsg is not menu-related
            // No break here !!!
  
        case WM_INITMENUPOPUP:
            if(g_pContextMenu2 != NULL)
                g_pContextMenu2->HandleMenuMsg(uMsg, wParam, lParam);

            // Inform caller that we handled WM_INITPOPUPMENU by ourself
            return(uMsg == WM_INITMENUPOPUP) ? 0 : TRUE;

        default:
            break;
    }

    // Call original WndProc of window to prevent undefined bevhaviour of window
    return ::CallWindowProc ((WNDPROC) GetProp ( hWnd, TEXT ("OldWndProc")), hWnd, uMsg, wParam, lParam);
}

int TShellContextMenu::GetPIDLCount(LPCITEMIDLIST pidl)
{
    BYTE * pCur = (BYTE *)pidl;
    int nCount = 0;

    if(pidl != NULL)
    {
        while(pidl->mkid.cb > 0)
        {
            pCur += pidl->mkid.cb;
            pidl  = (LPCITEMIDLIST)pCur;
            nCount++;
        }
    }
    return nCount;
}

UINT TShellContextMenu::GetPIDLSize(LPCITEMIDLIST pidl)
{  
    int nSize = 0;

    if(pidl != NULL)
    {
        while(pidl->mkid.cb != 0)
        {
            nSize += pidl->mkid.cb;
            pidl   = (LPITEMIDLIST)((LPBYTE)pidl + pidl->mkid.cb);
        }
    }
    return nSize;
}

void TShellContextMenu::FreePIDLList(LPITEMIDLIST * ppItemIDList, DWORD nItems)
{
    if(ppItemIDList != NULL)
    {
        for(DWORD i = 0; i < nItems; i++)
        {
            if(ppItemIDList[i] != NULL)
                HeapFree(m_hHeap, 0, ppItemIDList[i]);
        }

        HeapFree(m_hHeap, 0, ppItemIDList);
    }
}

LPBYTE TShellContextMenu::GetPIDLPos(LPCITEMIDLIST pidl, int nPos)
{
    BYTE * pCur = (BYTE *)pidl;
    int nCount = 0;

    if (pidl != NULL)
    {
        while(pidl->mkid.cb != 0)
        {
            if (nCount == nPos)
                return pCur;

            nCount++;
            pCur += pidl->mkid.cb;
            pidl  = (LPCITEMIDLIST)pCur;
        }
        if (nCount == nPos) 
            return pCur;
    }
    return NULL;
}

LPITEMIDLIST TShellContextMenu::CopyPIDL(LPCITEMIDLIST pidl, INT_PTR cb)
{
    LPITEMIDLIST pidlRet = NULL;

    if (cb == -1)
        cb = GetPIDLSize (pidl); // Calculate size of list.

    pidlRet = (LPITEMIDLIST)HeapAlloc(m_hHeap, HEAP_ZERO_MEMORY, cb + sizeof(USHORT));
    if(pidlRet != NULL)
        CopyMemory(pidlRet, pidl, cb);

    return pidlRet;
}

int TShellContextMenu::GetIContextMenu(HWND hWndParent, IContextMenu ** ppContextMenu, DWORD * pdwVersion)
{
    IContextMenu * pContextMenu = NULL;
    int nError = ERROR_SUCCESS;

    *ppContextMenu = NULL;

    // First we retrieve the normal IContextMenu interface
    // (every object should have it)
    if(nError == ERROR_SUCCESS)
    {
        nError = m_pShellFolder->GetUIObjectOf(hWndParent,
                                               m_dwFiles,
                                               (LPCITEMIDLIST *)m_ppItemIDList,
                                               IID_IContextMenu,
                                               NULL,
                                               (void **)&pContextMenu);
    }

    // Now try to get the higher version
    if(nError == ERROR_SUCCESS)
    {
        // Try to get the version 3
        if(pContextMenu->QueryInterface(IID_IContextMenu3, (void **)ppContextMenu) == S_OK)
        {
            pContextMenu->Release();
            pContextMenu = NULL;
            *pdwVersion = 3;
            return nError;
        }

        // Try to get the version 2
        if(pContextMenu->QueryInterface(IID_IContextMenu2, (void **)ppContextMenu) == S_OK)
        {
            pContextMenu->Release();
            pContextMenu = NULL;
            *pdwVersion = 2;
            return nError;
        }

        // So, we have to use version 1
        *ppContextMenu = pContextMenu;
        *pdwVersion = 1;
    }
   
    return nError;
}

// this is workaround function for the Shell API Function SHBindToParent
// SHBindToParent is not available under Win95/98
HRESULT TShellContextMenu::SHBindToParentEx(LPCITEMIDLIST pidl, REFIID riid, VOID ** ppv, LPCITEMIDLIST * ppidlLast)
{
    IShellFolder * pDesktopFolder = NULL;
    HRESULT hr = 0;
    int nCount;

    // If no pidl, desktop pidl or invalid pidl
    if(pidl == NULL || ppv == NULL || (nCount = GetPIDLCount(pidl)) == 0)
        return E_POINTER;

    SHGetDesktopFolder(&pDesktopFolder);

    if (nCount == 1)    // desktop pidl
    {
        if((hr = pDesktopFolder->QueryInterface(riid, ppv)) == S_OK)
        {
            if(ppidlLast) 
                *ppidlLast = CopyPIDL(pidl);
        }
    }
    else
    {
        IShellFolder * pShellFolder = NULL;
        LPITEMIDLIST pidlParent = NULL;
        LPBYTE pRel = GetPIDLPos(pidl, nCount - 1);

        pidlParent = CopyPIDL(pidl, pRel - (LPBYTE)pidl);
        
        hr = pDesktopFolder->BindToObject(pidlParent, NULL, IID_IShellFolder, (void **)&pShellFolder);
        if(hr == S_OK)
        {
            if ((hr = pShellFolder->QueryInterface(riid, ppv)) == S_OK)
            {
                if(ppidlLast != NULL)
                    *ppidlLast = CopyPIDL ((LPCITEMIDLIST) pRel);
            }
        }
        HeapFree(m_hHeap, 0, pidlParent);
        pShellFolder->Release ();
    }
   
    pDesktopFolder->Release ();
    return hr;
}

