/*****************************************************************************/
/* UrlButtons.cpp                         Copyright (c) Ladislav Zezula 2004 */
/*---------------------------------------------------------------------------*/
/* Functions related to URL buttons                                          */
/*---------------------------------------------------------------------------*/
/*   Date    Ver   Who  Comment                                              */
/* --------  ----  ---  -------                                              */
/* 30.05.04  1.00  Lad  The first version of UrlButtons.cpp                  */
/*****************************************************************************/

#define __UTILS_SELF
#include "Utils.h"
#include <assert.h>

//-----------------------------------------------------------------------------
// Local variables

static WNDPROC OldWindowProc;
static HCURSOR hCursor = NULL;
static UINT nIDUrlPoint = 0;
static UINT nIDUrlWait  = 0;

//-----------------------------------------------------------------------------
// Local functions

static LRESULT CALLBACK SetCursorProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    if(msg == WM_SETCURSOR && hCursor != NULL)
    {
        SetCursor(hCursor);
        return 0;
    }
    return CallWindowProc(OldWindowProc, hwnd, msg, wParam, lParam);
}

//-----------------------------------------------------------------------------
// Public functions

void SetURLCursors(UINT nIDUrlPoint, UINT nIDUrlWait)
{
    ::nIDUrlPoint = nIDUrlPoint;
    ::nIDUrlWait  = nIDUrlWait;
}

// This function sets the font for one URL button.
// They must be a button class with owner-draw style set.
int InitURLButton(HWND hDlg, UINT nIDCtrl, BOOL bBigFont)
{
    LOGFONT logFont;
    HFONT   hFont = NULL;
    HWND    hWnd = GetDlgItem(hDlg, nIDCtrl);

    // If the button is not there, do nothing.
    if(hWnd == NULL)
        return ERROR_FILE_NOT_FOUND;

    // Retrieve the settings of the dialog font.
    // The buttons should have the same font like the dialog font.
    hFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0);
    GetObject(hFont, sizeof(LOGFONT), &logFont);

    // Setup the font
    logFont.lfUnderline = TRUE;
    if(bBigFont != FALSE)
    {
        logFont.lfHeight = 18;
        logFont.lfWeight = FW_BOLD;
    }

    // Create the font
    hFont = CreateFontIndirect(&logFont);
    if(hFont == NULL)
        return GetLastError();

    // Load URL-Point cursor for this buttons
    hCursor = LoadCursor(g_hInst, MAKEINTRESOURCE(nIDUrlPoint));
    if(hCursor == NULL)
        return GetLastError();

    if(IsURLButton(hWnd))
    {
        // Set the new window font and procedure
        SendMessage(hWnd, WM_SETFONT, (LPARAM)hFont, 0);
        OldWindowProc = (WNDPROC)SetWindowLongPtr(hWnd, GWLP_WNDPROC, (LONG_PTR)SetCursorProc);
    }
    return ERROR_SUCCESS;
}

// This function sets the font for the URL buttons.
// They must be a button class with owner-draw style set.
int InitURLButtons(HWND hDlg)
{
    LOGFONT logFont;
    HFONT   hFont = NULL;
    HWND    hWnd = GetWindow(hDlg, GW_CHILD);
    
    // Retrieve the settings of the dialog font.
    // The buttons should have the same font like the dialog font.
    hFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0);
    GetObject(hFont, sizeof(LOGFONT), &logFont);
    logFont.lfUnderline = TRUE;
    hFont = CreateFontIndirect(&logFont);
    if(hFont == NULL)
        return GetLastError();

    // The URL Point cursor must be set
    assert(nIDUrlPoint != 0);

    // Load URL-Point cursor for this buttons
    hCursor = LoadCursor(g_hInst, MAKEINTRESOURCE(nIDUrlPoint));
    if(hCursor == NULL)
        return GetLastError();

    while(hWnd != NULL)
    {
        if(IsURLButton(hWnd))
        {
            // Set the new window font and procedure
            SendMessage(hWnd, WM_SETFONT, (LPARAM)hFont, 0);
            OldWindowProc = (WNDPROC)SetWindowLongPtr(hWnd, GWLP_WNDPROC, (LONG_PTR)SetCursorProc);
        }
        
        // Get the next window
        hWnd = GetWindow(hWnd, GW_HWNDNEXT);
    }
    return ERROR_SUCCESS;
}

BOOL IsURLButton(HWND hWnd)
{
    TCHAR szClassName[0x20] = _T("");
    DWORD dwStyle;
    
    GetClassName(hWnd, szClassName, _tsize(szClassName));
    dwStyle = GetWindowLong(hWnd, GWL_STYLE);

    // Must be a button
    if(_tcsicmp(szClassName, _T("Button")))
        return FALSE;
        
    // Must not be a groupbox
    if((dwStyle & BS_GROUPBOX) == BS_GROUPBOX)
        return FALSE;

    // If ownerdrawn, the OK.
    return ((dwStyle & BS_OWNERDRAW) == BS_OWNERDRAW) ? TRUE : FALSE;
}

// Draws hypertext button. It is drawn with the blue color
// when normal and with the red color when clicked.
BOOL DrawURLButton(HWND hDlg, LPDRAWITEMSTRUCT dis, BOOL bCentered)
{
    TCHAR szText[240];
    RECT  rect;
    UINT  nLength = 0;
    HWND  hWnd = GetDlgItem(hDlg, dis->CtlID);
    SIZE  size;
    int   x = 0;
    int   y = 0;

    if(dis == NULL)
        return FALSE;

	if(dis->CtlType == ODT_BUTTON)
	{
        // Retrieve text from button
        nLength = GetWindowText(hWnd, szText, _tsize(szText));
		
        // Select color and draw it.
        if(dis->itemState & ODS_SELECTED)
			SetTextColor(dis->hDC, RGB(0xFF, 0x00, 0x00));
        else
			SetTextColor(dis->hDC, RGB(0x00, 0x00, 0xFF));
		
        // If the text has to be centered, calculate the position
        // where to paint it
        if(bCentered)
        {
            GetClientRect(hWnd, &rect);
            GetTextExtentPoint32(dis->hDC, szText, nLength, &size);

            x = (rect.right  - size.cx) / 2;
            y = (rect.bottom - size.cy) / 2;
        }
        
        TextOut(dis->hDC, x, y, szText, nLength);
	}

    return TRUE;
}

// This callback handles the mouse click on an URL button (WWW, mail, http).
// It invokes the default web browser/mail client for the button
// (Recognized from button text)
BOOL ClickURLButton(HWND hButton)
{
    HCURSOR hWaitCursor = LoadCursor(g_hInst, MAKEINTRESOURCE(nIDUrlWait));
    TCHAR   szUrl[128] = _T("");
    TCHAR   szText[128];

    // This happens when the SetURLCursors
    // has not been called in WinMain/DllMain
    assert(nIDUrlPoint != 0);
    assert(nIDUrlWait != 0);

    // Retrieve button text
    GetWindowText(hButton, szText, _tsize(szText));

    // Test E-mail addresses. They must have "mailto:" at the begin
    if(_tcschr(szText, _T('@')) != NULL)
    {
        _tcscpy(szUrl, _T("mailto:"));
        _tcscat(szUrl, szText);
    }

    // Test "http://"
    if(!_tcsnicmp(szText, _T("http://"), 7))
        _tcscpy(szUrl, szText);

    // If only "www." at the begin, we have to add "http://"
    if(!_tcsnicmp(szText, _T("www."), 4))
    {
        _tcscpy(szUrl, _T("http://"));
        _tcscat(szUrl, szText);
    }

    // If any valid URL address, invoke the default application
    // Set waiting cursor and launch default browser
    if(szUrl[0] != 0)
    {
        HCURSOR hOldCursor = SetCursor(hWaitCursor);

        ShellExecute(NULL, _T("open"), szUrl, NULL, NULL, SW_SHOWNORMAL);
        SetCursor(hOldCursor);
    }
    return TRUE;
}
